"""

TorpigSpy - Torpig/MBR Configuration Decryption
Michael Hale Ligh and Greg Sinclair @ iDefense
http://code.google.com/mhl-malware-scripts

"""

from immlib import LogBpHook
import immlib

ProgName="TorpigSpy"

# Ensure the HANDLE passed to API functions is for the named pipe
def isTheRightHandle (imm, hfile):
        handles = imm.getAllHandles()
        for h in handles.keys():
            han = handles[h]
            if (han.getHandle() == hfile) and (han.getNativeName() == "\\Device\\NamedPipe\\!win$"):
                return True
        return False

class MyOwnHook(LogBpHook):
    def __init__(self):
        LogBpHook.__init__(self)

    def run(self,regs):
        imm = immlib.Debugger()
        function_name = imm.getKnowledge("%08x" % regs['EIP'])
        # If the created file is !win$, enable the read and write hooks
        if function_name == "CreateFileW":
            addr = imm.readLong(regs['ESP'] + 0x4)
            filename = imm.readWString(addr)
            filename = filename.replace("\x00", "")
            imm.Log("%s: Opening %s" % (ProgName, filename))
            if filename == "\\\\.\\pipe\\!win$":
                ReadFile    = imm.getAddress("kernel32.ReadFile")
                WriteFile   = imm.getAddress("kernel32.WriteFile")
                self.add("readfile",  ReadFile)
                self.add("writefile", WriteFile)
                imm.addKnowledge("%08x" % ReadFile,  "ReadFile",  force_add=0x1)
                imm.addKnowledge("%08x" % WriteFile, "WriteFile", force_add=0x1)
                imm.removeHook("createfile")
                imm.Log("%s: Set ReadFile (0x%x) and WriteFile (0x%x) Hook" % (ProgName, ReadFile, WriteFile))
        # Capture and analyze the data received from the named pipe
        elif function_name == "ReadFile":
            hfile = imm.readLong(regs['ESP'] + 0x4)
            if isTheRightHandle(imm,hfile) == True:
                size = imm.readLong(regs['ESP'] + 0xC) # DWORD
                desc = ""
                # Determine the purpose of the received data
                if size == 1:
                    desc = "< Request Ack"
                elif size == 4:
                    desc = "< Response size"
                else:
                    desc = "< Response "
                log_items = ["Read", "0x%x" % size, "N/A", "%s" % desc]
                torpig_spy = imm.getKnowledge("torpig_spy")
                torpig_spy.add(regs['EIP'], log_items)
        # Capture and analyze the data sent to the named pipe
        elif function_name == "WriteFile":
            hfile = imm.readLong(regs['ESP'] + 0x4)
            if isTheRightHandle(imm,hfile) == True:
                addr = imm.readLong(regs['ESP'] + 0x8)
                size = imm.readLong(regs['ESP'] + 0xC)
                buf = imm.readMemory(addr, size)
                strbuf = ""
                for i in range (0,size):
                    strbuf += "%2.2x " % ord(buf[i])
                desc = ""
                # Determine the purpose of the sent data
                if (size == 1) and (ord(buf[0]) == 0x1):
                    desc = "> Client Hello 1"
                elif (size == 1) and (ord(buf[0]) == 0x2):
                    desc = "> Client Hello 2"
                elif (size == 4):
                    desc = "> Request size (%d bytes)" % imm.readLong(addr)
                elif (size == 6):
                    desc = "> Request structure (Type %d)" % ord(buf[4])
                log_items = ["Write", "0x%x" % size, "%s" % strbuf, "%s" % desc]
                torpig_spy = imm.getKnowledge("torpig_spy")
                torpig_spy.add(regs['EIP'], log_items)

def main():
    imm = immlib.Debugger()
    imm.ignoreSingleStep("CONTINUE")

    # Set up the output window
    column_titles = ["Hook", "Size", "Data", "Description"]
    torpig_spy = imm.createWindow("Torpig Spy", column_titles)
    imm.addKnowledge("torpig_spy", torpig_spy, force_add=0x1)

    # Place a hook on CreateFileW
    CreateFileW = imm.getAddress("kernel32.CreateFileW")
    myhooks = MyOwnHook()
    myhooks.add("createfile",CreateFileW)
    imm.addKnowledge("%08x" % CreateFileW, "CreateFileW", force_add=0x1)

    imm.Log("%s: Started!" % ProgName)